/*
********************************************************************************
**  hgc text16 operators
********************************************************************************
**
   Copyright (C) 1997      Jochen Hoenicke, 
**			    [jochen.hoenicke@informatik.uni-oldenburg.de]
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**
*/

#ifndef _HGC_TEXT16_INC_
#define _HGC_TEXT16_INC_

#define	MAINTAINER	"Jochen Hoenicke, [jochen.hoenicke@informatik.uni-oldenburg.de]"

/*
**     GGI includes
*/
#include "IBM/vga_text16.h"

#define POINTER_MASK_AND 0xFFFF		/* Attribute, Character */
#define POINTER_MASK_XOR 0x7700

#ifndef CURSOR_HW
#define CURSOR_MASK_AND	 0xFFFF		/* Attribute, Character */
#define CURSOR_MASK_XOR	 0x7700
#endif

typedef struct text16_point_s {
	int		visible;	/* Should I visible the old data? */
	int		offset;		/* Offset of old point */
	charcell	data;		/* Data at old point */
} text16_point;
	
struct text16_data_s {
	kgi_uint		magic;
	unsigned long	origin;
	text16_point	pointer;
#ifndef CURSOR_HW
	text16_point	cursor;
#endif
};

static struct text16_data_s text16_data[MAX_NR_DISPLAYS];

#define ORIGIN 		(data->origin)

#define PVISIBLE	(point->visible)
#define POFFSET		(point->offset)
#define PDATA		(point->data)

static void text16_data_init(void)
{
	memset(text16_data,0,sizeof(struct text16_data *)*MAX_NR_DISPLAYS);
}
	
static struct text16_data_s *text16_data_realloc(struct text16_data_s *old)
{
	int i;

	if (old==NULL || old->magic!=TEXT16_MAGIC) {
		for (i=0;i<MAX_NR_DISPLAYS;i++) 
			if (text16_data[i].magic!=TEXT16_MAGIC) {
			  memset(&(text16_data[i]),0,sizeof(struct text16_data_s));
			  text16_data[i].magic=TEXT16_MAGIC;
			  return &(text16_data[i]);
			}
		return NULL;
	}
	memset(old,0,sizeof(struct text16_data_s));
	old->magic=TEXT16_MAGIC;
	return old;
}

/* With 4096 byte frame buffer this doesn't make much sense */
static void text16_set_origin(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
	IS_CRITICAL;
	register kgi_uint xpix = x / dpy->mode.request.text.x;
	register kgi_uint ypix = y / dpy->mode.request.text.y;

	register kgi_uint offset = xpix + ypix*dpy->fb.size.x;

	TRACE(printk(KERN_DEBUG "Origin: 0x%.8x -> 0x%.8x\n",ORIGIN,offset));

	ENTER_CRITICAL;

		if (ORIGIN != offset) {

			outbCRT(offset >> 8, CRT_STARTADDR_H);
			outbCRT(offset,      CRT_STARTADDR_L);
			ORIGIN = offset;
		}

	LEAVE_CRITICAL;
}

#define text16_set_split	NULL	/* !!! implement! */

#define	text16_set_frame	NULL	/* !!! implement! */

#define TEXTX           (dpy->mode.request.text.x)
#define TEXTY           (dpy->mode.request.text.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

static inline void text16_show_point(struct kgi_display *dpy, kgi_uint x, kgi_uint y, 
				     text16_point *point,
				     charcell and, charcell xor) 
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
        register kgiu32 newoff =ORIGIN + (x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
        register charcell *p = (void *) dpy->fb.pbuf;
	register charcell cell;

 	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	if (PVISIBLE) {
		if (newoff != POFFSET)
			return;		/* Nothing to do */
		else
			*(p+POFFSET)=PDATA;
	}

	cell= *(p+newoff);
	*(p+newoff) = (cell & and) ^ xor;

	PVISIBLE=1;
	POFFSET=newoff;
	PDATA=cell;
}	

static inline void text16_hide_point(struct kgi_display *dpy,
				     text16_point *point ) 
{
	if (PVISIBLE) {
		register charcell *p = (void *) dpy->fb.pbuf;

		*(p+POFFSET)=PDATA;
		PVISIBLE=0;
	}
}
	
void text16_show_pointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	text16_show_point(dpy, x, y, 
		&((struct text16_data_s *)(dpy->driver_data))->pointer, 
		POINTER_MASK_AND, POINTER_MASK_XOR);
}

static void text16_hide_pointer(struct kgi_display *dpy)
{
	text16_hide_point(dpy,
		&((struct text16_data_s *)(dpy->driver_data))->pointer); 
}

#define text16_undo_pointer     text16_hide_pointer

#ifndef CURSOR_HW
static void text16_show_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	text16_show_point(dpy, x, y, 
		&((struct text16_data_s *)(dpy->driver_data))->cursor,
		CURSOR_MASK_AND, CURSOR_MASK_XOR);
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	text16_hide_point(dpy,
		&((struct text16_data_s *)(dpy->driver_data))->cursor);
}

#define text16_undo_cursor      text16_hide_cursor

#else

static void text16_show_cursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	struct text16_data_s *data = (struct text16_data_s *)(dpy->driver_data);
        register kgiu32 new = ORIGIN + (x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
 	IS_CRITICAL;

        ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	ENTER_CRITICAL;
	outbCRT(new >> 8, CRT_CURSORADDR_H);
	outbCRT(new&0xFF, CRT_CURSORADDR_L);
	LEAVE_CRITICAL;
}

static void text16_hide_cursor(struct kgi_display *dpy)
{
	IS_CRITICAL;

	ENTER_CRITICAL;
	outbCRT(0xFF,CRT_CURSORADDR_H);
	outbCRT(0xFF,CRT_CURSORADDR_L);
	LEAVE_CRITICAL;
}

#define text16_undo_cursor      text16_hide_cursor
#endif /* CURSOR_HW */

#undef SPLITLINE
#undef ORIGIN

#undef PVISIBLE
#undef POFFSET
#undef PDATA

#endif /* _HGC_TEXT16_INC */
