/* $Id: sdac.c,v 1.10 1998/11/24 18:17:13 marcus Exp $
******************************************************************************

   S3 SDAC/Trio32/Trio64 clock chip driver

   Copyright (C) 1995 by Steffen Seeger [seeger@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

	This code was derived from the following source of information:
	[1]	80C716 SDAC Data Book (c) 1994 by S3 Incorporated
	[2]	Trio32/Trio64 Data Book (c) 1994 by S3 Incorporated

******************************************************************************
*/

#warning	not converted to 0.0.9 yet.

#define	MAINTAINER	"Steffen Seeger, seeger@ggi-project.org"

#include <kgi/module.h>

#ifdef TRIO
  #include "S3/Trio.h"

  static struct kgi_clock clock =
  {
	"S3 Incorporated",	/* manufacturer		*/
	"Trio32/Trio64",	/* model		*/
	GM_ALL,			/* modes		*/ 
	0,			/* flags		*/
	{ 0x7FFF, 0x7FFF },	/* maximum resolution	*/
	NULL,			/* Private data hook	*/
	CLK_TYPE_PROG,		/* clock type		*/
	{{ 25000000, 110000000,	/* clock ranges		*/
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
                  0,         0}}
  };
#else
  #include "S3/sdac.h"

  static struct kgi_clock clock =
  {
	"S3 Incorporated",	/* manufacturer		*/ 
	"80C716-ME/MG (SDAC)",	/* model		*/
	GM_ALL,			/* supported modes 	*/
	0,			/* special capabilities	*/
	{ 0x7FFF, 0x7FFF },	/* maximum resolution	*/
	NULL,			/* Private data hook	*/
	CLK_TYPE_PROG,		/* clock type		*/
	{{   400000, 119000000,	/* clock ranges		*/
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
	          0,         0,
                  0,         0}}
  };
#endif

#define	MIN_FREQ clock.clock.range[0].min
#define	MAX_FREQ clock.clock.range[0].max

/*
	Clock driver dependend #define's
*/

#define CLK_VERSION "0.00 ALPHA"

#define	fref	14318180   /* Hz, must be within 14..15 MHz */ 

#define F(M,N,FVCO)  (((M)+2)*(FVCO)/((N)+2))

#define	INVALIDCLK 0xFFFF

static kgiu16 clk_last_cmd = 0xFFFF;
static kgi_sint    clk_last_freq = -1;

static void setPLL(kgiu16 cmd, kgi_sint clk)
{
  #ifdef TRIO
    #warning implement setPLL !!!
  #else
    outbDAC(clk ,DAC_PLL_W_INDEX);	/* select clock generator	*/
    outbDAC(cmd & 0xFF, DAC_PLL_DATA);	/* 1st byte			*/
    outbDAC(cmd >> 8, DAC_PLL_DATA);	/* 2nd byte			*/
  #endif
}

static kgiu16 calcfreq(int *freq, enum kgi_tm_cmd cmd)
{
  int fvco = *freq;
  int hf   = *freq;
  int div  = 0;
  int n;
  struct { kgiu8 m; kgiu8 n; int delta; }
         best = { 0xFF, 0xFF, 0x7FFFFFFF };    
 
  if(*freq == clk_last_freq)	/* we never get called with freq == -1 */
    return clk_last_cmd;
  
  while((div < 4) && (hf < 135000000))
    { div += 1; hf += hf; }
 
  if((hf < 135000000) || (270000000 < hf) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz)\n",__FILE__,__LINE__,\
                   *freq));
      return INVALIDCLK;
    }

  fvco  = fref >> div;

  for(n = 1; n < 32; n++)
    { 
      register int m = hf*(n+2)/fref - 2;
      register int delta = *freq - F(m,n,fvco);

      delta = (cmd == CMD_LOWER) ? delta : -delta;

      if(m > 127) break;
      if((delta < 0) || (delta > best.delta)) continue;

      best.n = n; best.m = m; best.delta = delta;
    }

  if((best.n > 31) || (best.m > 127) || (div > 3))
    {
      TRACE(printk("%s:%i: invalid clock request (%i Hz), n=%i, m=%i, d=%i\n",
                   __FILE__,__LINE__, *freq, best.n, best.m, div)); 
      return INVALIDCLK; 
    }
   else
    {
      TRACE(printk("%s: request for DCLK = %i Hz,", __FILE__,*freq ));
      *freq = clk_last_freq = F(best.m,best.n,fref >> div);
      TRACE(printk(" did %i Hz.\n", *freq));
      return clk_last_cmd = best.m | (best.n << 8) | (div << 13);
    }
}

/*
     KGI module interface
*/

static kgiu16 PLL_save[4];
static kgiu8 clk, PLL_ctrl;

struct kgi_clock *clock_init(void)
{
  /*
  	Detection is done by reading CLK0 and CLK1 frequencies. As these are
  	read only it should be save enough. (For the SDAC, the Trio will be 
  	detected by the chipset driver.)
  */

  #ifdef TRIO
    #warning implement kgi_clock_init !!!	
  #else
    clk = getCLK();
    outbDAC(0, DAC_PLL_R_INDEX);		/* save PLL registers 0 - 15 */
    insbDAC(DAC_PLL_R_INDEX, &PLL_save, 8);
    /*
    if((PLL_save[0] != ...) || (PLL_save[1] != ...))
      return FALSE;
    */
    outbDAC(DAC_PLL_CONTROL, DAC_PLL_R_INDEX);
    PLL_ctrl = inbDAC(DAC_PLL_DATA);
  #endif
  
  printk("%s %s clock driver v"CLK_VERSION".\n", clock.manufact, clock.model);
  return &clock;
}

void clock_done(void)
{
  #ifdef TRIO
    #warning implement kgi_clock_done !!!
  #else
    setCLK(clk);
    setPLL(PLL_save[3], 3);
    outbDAC(DAC_PLL_CONTROL, DAC_PLL_W_INDEX);  
    outbDAC(PLL_ctrl, DAC_PLL_DATA);
  #endif
  TRACE(printk("%s %s clock driver removed.\n", clock.manufact, clock.model));
}

kgi_sint clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			 enum kgi_tm_cmd *cmd)
{
  int newfreq = (*cmd == CMD_PROPOSE) 
                    ? 0 : mode->dclk * mode->clk.mul / mode->clk.div;

  switch (*cmd)
   {
     case CMD_PROPOSE:
	mode->clock = INVALIDCLK;
	return EOK;

     case CMD_LOWER:
	if (newfreq < MIN_FREQ)
	  {
	    TRACE(printk("%s: requested DCLK too low (%i Hz).\n", \
			 __FILE__, mode->dclk));
	    return -E(CLOCK, UNKNOWN);
	  }

   	if (newfreq > MAX_FREQ)	/* meet limits */
	  newfreq = MAX_FREQ;

	mode->clock = calcfreq(&newfreq, *cmd);
	mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

	return (mode->clock == INVALIDCLK) ? -E(CLOCK, UNKNOWN) : EOK;
			
     case CMD_RAISE:
	if (MAX_FREQ < newfreq)
	  {
	    TRACE(printk("%s: requested DCLK too high (%i Hz).\n", \
			 __FILE__, mode->dclk));
	    return -E(CLOCK, UNKNOWN);
	  }

	if (newfreq < MIN_FREQ)
	  newfreq = MIN_FREQ;

	mode->clock = calcfreq(&newfreq, *cmd);
	mode->dclk = newfreq * mode->clk.div / mode->clk.mul;

	return (mode->clock == INVALIDCLK) ? -E(CLOCK, UNKNOWN) : EOK;

     case CMD_CHECK:
	if ((newfreq < MIN_FREQ) || (newfreq > MAX_FREQ))
	  return -E(CLOCK, UNKNOWN);

	mode->clock = calcfreq(&newfreq, CMD_LOWER);

	if (   (mode->clock != INVALIDCLK)
	    && (   ((mode->dclk*mode->clk.mul/mode->clk.div) - newfreq) 
		 < ((mode->dclk*mode->clk.mul/mode->clk.div)*1/100)))
	  return EOK;

	return -E(CLOCK, UNKNOWN);

     default:
	INTERNAL_ERROR;
	return -E(CLOCK, UNKNOWN);
    }
}

void clock_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
  if(mode->clock != INVALIDCLK)
    {
      #ifdef TRIO
	#warning implement kgi_clock_set_mode !!
      #else
        /* always use CLK3, which is reprogrammed according to the request. */
        setPLL(mode->clock, 3);

	outbDAC(DAC_PLL_CONTROL, DAC_PLL_W_INDEX);
	outbDAC(0, DAC_PLL_DATA);

	setCLK(3);
      #endif
    }
   else
    { /* This should never happen !! */
      INTERNAL_ERROR;
    }
}

kgi_sint clock_ioctl(struct inode *inode, struct file *file, kgi_uint cmd,
	       kgiu32 arg)
{
  return -E(CLOCK, UNKNOWN);
}
